-- LibRealmDetect-1.0, by pankker / Maunotavast Zenedar-EU
-- LibRealmDetect-1.0 is hereby placed in the Public Domain.

local MAJOR, MINOR = "LibRealmDetect-1.0", 1

assert(LibStub,                           MAJOR.." requires LibStub")
assert(LibStub("CallbackHandler-1.0", 1), MAJOR.." requires CallbackHandler-1.0")
assert(LibStub("LibWho-2.0", 1),          MAJOR.." requires LibWho-2.0")

local lib = LibStub:NewLibrary(MAJOR, MINOR)
if not lib then return end

lib.wholib    = lib.wholib or LibStub("LibWho-2.0"):Library()
lib.callbacks = lib.callbacks or 
   LibStub("CallbackHandler-1.0"):New(lib, "RD_RegisterCallback", "RD_UnregisterCallback", false)

lib.WHOLIB_OPTS = 
   { queue = lib.wholib.WHOLIB_QUEUE_QUIET, 
     callback = "ParseWho", 
     handler = lib 
   }

lib.SEARCH_PLAN = 
   -- Use mapIDs for locale-independency. Only use zones with CRZ disabled but 
   -- cross-realm group phasing enabled (i.e. mostly just Pandaria).
   { Alliance = 
      { 905, -- Shrine of Seven Stars
        951, -- Timeless Isle
        807, -- Valley of the Four Winds
        806, -- The Jade Forest
        811, -- Vale of Eternal Blossoms
        809, -- Kun-Lai Summit
        928, -- Isle of Thunder
        810, -- Townlong Steppes
        858, -- Dread Wastes
        857, -- Krasarang Wilds
        873, -- The Veiled Stair
        929  -- Isle of Giants
      },      
     Horde = 
      { 903, -- Shrine of Two Moons
        951, -- Timeless Isle
        807, -- Valley of the Four Winds
        806, -- The Jade Forest
        811, -- Vale of Eternal Blossoms
        809, -- Kun-Lai Summit
        928, -- Isle of Thunder
        810, -- Townlong Steppes
        858, -- Dread Wastes
        857, -- Krasarang Wilds
        873, -- The Veiled Stair
        929  -- Isle of Giants
      } 
   }

lib.MATCH_THRESHOLDS = -- TODO: Maybe adjust these values a bit!
   { {25, 0.5 },       -- for 25 and more players searched, at least 50% need to be from the same realm / realmgroup
     {10, 0.75},       -- for 24-10 players, at least 75%
     {5 , 0.9 }        -- for 9-5, at least 90%. 
   }                   -- More zones will be searched if none of these are matched after a /who search.

lib.KNOWN_CONNECTED_REALMS = 
   {
      -- ["Zenedar-EU"] = "Zenedar / Bladefist",
      -- ["Bladefist-EU"] = "Zenedar / Bladefist"
      -- TODO: fill in known realm connections. Also, probably, well, DO need to somehow detect the region used by the game.
   }

local _G               = _G
local REALM_SEPARATORS = _G.REALM_SEPARATORS
local WHO_TAG_ZONE     = _G.WHO_TAG_ZONE
local UnitFactionGroup = _G.UnitFactionGroup
local GetRealmName     = _G.GetRealmName

-----------------------------------
-- External functions
-----------------------------------
function lib:GetCurrentRealm()
   return lib.currentRealm, lib.realmUncertain
end

function lib:GetHomeRealm()
   return lib.homeRealm
end

function lib:GetRealmTimeZone()
   -- TODO: do this
end

function lib:GetRealmTime()
   -- TODO: do that
end

-----------------------------------
-- Embed / External
-----------------------------------
lib.EXTERNALS = { "GetCurrentRealm", 
                  "GetHomeRealm",
                  "GetRealmTimeZone",
                  "GetRealmTime",
                  "RD_RegisterCallback",
                  "RD_UnregisterCallback" }

lib.embeds = lib.embeds or {}

function lib:Embed(handler)
   for k,v in ipairs(lib.EXTERNALS) do
      handler[v] = lib[v]
   end
   
   lib.embeds[handler] = 1
   
   return handler
end

function lib:Library()
   return lib:Embed({})
end

-----------------------------------
-- Internal functions
-----------------------------------
function lib:StopDetect()
   lib.detecting = nil
   lib.matchResults = nil
end

function lib:DetectCurrentRealm(ongoingDetect)
   if lib.detecting and not ongoingDetect then return end -- trying to initiate detect with another detect already in progress
   
   local faction = UnitFactionGroup("player")
   local plan = lib.SEARCH_PLAN
   plan = plan and plan[faction]
   if not plan then
      -- player has no faction, do not try to detect realm through /who.
      if not lib.nofactionFired then
         lib.callbacks:Fire("REALMDETECT_ERROR_NO_FACTION")
      end
      lib.nofactionFired = 1
      return 
   end
   
   if not ongoingDetect then
      lib.detecting = 1
      lib.matchResults = { searchPlan = plan, queryCount = 0, 
                           playerCount = 0, players = {}, realms = {} }
   else
      if not lib.matchResults then 
         -- a non-ongoing detect claiming to be ongoing
         lib.detecting = nil
      end
   end
   
   local queryCount = lib.matchResults.queryCount + 1
   lib.matchResults.queryCount = queryCount
   
   local mapID = lib.matchResults.searchPlan[queryCount]
   if not mapID then 
      -- end of search plan reached. Or something just broke ...
      lib.detecting = nil
      lib.callbacks:Fire("REALMDETECT_ERROR_NOT_ENOUGH_PLAYERS_FOUND")
      return
   end
   local searchString = ('%s"%s"'):format(WHO_TAG_ZONE, GetMapNameByID(mapID))
   
   lib.wholib:Who(searchString, lib.WHOLIB_OPTS)
end

function lib:ParseWho(query, results, complete)
   local mr = lib.matchResults
   
   if not mr then
      lib.detecting = nil
      return 
   end
   
   for _, player in ipairs(results) do
      if not mr.players[player.Name] then 
         -- keep track of previously matched players if doing multiple queries
         -- (only count them once)
         mr.players[player.Name] = 1
         
         local name, realm = strsplit(REALM_SEPARATORS, player.Name)
         realm = realm or lib:GetHomeRealm()
         
         mr.realms[realm] = (mr.realms[realm] or 0) + 1
         mr.playerCount = mr.playerCount + 1
      end  
   end
   
   local winnerRealm = lib:GetBestRealmCandidate(mr)
   
   if winnerRealm then 
      lib:OnRealmKnown(winnerRealm)
   else
      lib:DetectCurrentRealm(1) 
   end
end

function lib:GetBestRealmCandidate(results)
   local total = results.playerCount
   
   local highestRealmTotal = 0
   local highestRealmName
   
   -- get realm with highest amount of players found.
   for k,v in pairs(results.realms) do
      if v > highestRealmTotal then
         highestRealmName = k
         highestRealmTotal = v
      end
   end
   
   local thresholds = lib.MATCH_THRESHOLDS
   
   -- ... and check if it matches one of the thresholds
   for k,v in ipairs(thresholds) do
      local minTotal, minPct = v[1], v[2]
      if total >= minTotal then
         local highestPct = (highestRealmTotal / total)
         return (highestPct >= minPct) and highestRealmName
      end
   end
end

function lib:OnRealmKnown(realm)
   local hasChanged = tostring(realm) ~= tostring(lib.currentRealm)
   lib.currentRealm = realm
   lib.realmUncertain = nil
   lib.detecting = nil
   lib.initialDetect = nil
   lib.callbacks:Fire("REALMDETECT_REALM_KNOWN", lib.currentRealm, hasChanged)
end

function lib:OnRealmChangeDetected()
   lib:StopDetect()
   if (not InCombatLockdown() and not IsInGroup()) then
      lib:OnRealmKnown(lib.homeRealm)
   else
      lib.realmUncertain = 1
      lib.callbacks:Fire("REALMDETECT_REALM_UNKNOWN", lib.currentRealm)
      lib:DetectCurrentRealm()
   end
   
end

-----------------------------------
-- Frame and event setup
-----------------------------------
do
   if not lib.frame then
      lib.frame = CreateFrame("Frame")
   end
   
   local f = lib.frame
   f.events = {}
   
   function f.events:PLAYER_ENTERING_WORLD()
      if not lib.currentRealm then
         lib.homeRealm = GetRealmName()
         lib.initialDetect = 1
         lib:OnRealmChangeDetected()
      end
   end
   
   function f.events:UNIT_SPELLCAST_SUCCEEDED(unit, spell, rank, lineid, spellid)
      if unit == "player" and spellid == 836 and not lib.initialDetect then -- 836 = LOGINEFFECT
         lib:OnRealmChangeDetected()
      end
   end
   
   -- TODO: detect realm change when joining a cross-realm group 
   --       in a CRZ already shared by you and the group, or in non-phasing 
   --       zones (SW, Orgrimmar), where no phasing takes place immediately.
   
   f:UnregisterAllEvents()
   for k,v in pairs(f.events) do
      f:RegisterEvent(k)
   end
   
   f:SetScript("OnEvent", function(self, event, ...) self.events[event](self, ...) end)
end

-----------------------------------
-- Re-embed
-----------------------------------
for target, _ in pairs(lib.embeds) do
	lib:Embed(target)
end